//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //   
//  These routines are intended to interface with EDICT, which is    //
//  a Japanese/English Dictionary developed and copyrighted by       //
//  James William Breen.                                             //
//                                                                   //   
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  This module implements the dictionary search and mantinance routines.
//  These handle all interactions with the dictionaries and with searching.
//
//  The main dicitonary procedure involves searching Jim Breen's Japanese-
//  English Dictionary based on the index file used by JWP.  
//
//  Understanding the index file:
//
//  The dictionary is index by a string based on a EUC character 
//  combination.  The index is based on ascii characters of three 
//  or more chracters, kana strings of 2 or more kana, and single 
//  kanji.  All of the indexs are mixxed togetther and sorted in
//  numberical order, allowing use of a binary search.  Each index
//  points to a match string location in the dictionary file.  Note 
//  that these matches may actually be in the middle of a dictionary
//  entry.
//
//  The index has some odd characteristics.  Kanji and ascii are 
//  indexed wherever they occure in the text, but kara strings are only
//  indexed when they begin a word!
//
//  A search is conducted by performing a binary search on the index.  
//  For each index value we get the line from the actual dictionary and
//  compare with our key.  When we find the first match that matches 
//  our key, we have the beginning of our actuall data.  We then containue
//  getting the next index, and then the next data from the dicitionary,
//  until we get to the point when the lines gotten do not match the key.
//
//  Understanding dicitioanry entries:
//
//  Each entry inthe dictionary is a EUC line of text, terminated by a
//  cr-lf pair. (I do not count on this actually being a cr-lf, and can 
//  handle any comination [cr, lf, lf-cr, cr-lf].).  The format of each
//  line depends on weather or not there is kanji for the word in 
//  in quetion:
//
//      <kanji> [<kana>] /<defintion>/<definition>/ ...
//
//              or if there is no kanji:
//
//      <kana> /<definition>/<definition>
//
//  The index seach will give us the location of what we are searching 
//  for within a line of text (definititely not necessarily at the 
//  actual beginning of the line).  To handle this problem i use a 
//  centered buffer approche (see below).  Generally, we backup to the 
//  beginning of the line.  
//
//  Each of the <definition>s contains a complete definition of the 
//  word.  Since words are offten used to mean different things, these
//  defintions can vary a lot.  Within each definition, however, there
//  may be key string.  This string:
//
//          (<key>,<key>,...)
//
//  Tells us soemthing about the type of defintiion.  For example, the 
//  string (pl,pn,giv,fem) tells us that this entry is a place name, and
//  a female given name.  
//
//  I allow selective searching based on the type of entry.  The main 
//  desire of this is to allow filtering of personal names, and place 
//  names.  For example searching on the kanji NI (two) you get 738 
//  entries.  Removing place names and personal names, reduces this to
//  277 entries.  
//
//  Filtering keys correctly took a lot of work.  See the routine 
//  JFC_dict::search_dict() for a detailed descritption of the 
//  filtering process.
//
//  The dominate keys are:
//
//      DICTKEY_END    -- Results must match the end of a word.
//      DICTKEY_BEGIN  -- Reuslts must match the beginning of a word.
//      DICTKEY_NAMES  -- Rejects all personal names.
//      DICTKEY_PLACES -- Rejects all place names.
//
//  User dictionary:
//
//  The user dictionary is kept in the same format as EDICT.  When the 
//  dictionary is loaded a '\n' character is placed in front of the 
//  actual dictionary.  This allows the search backward in the file to
//  correctly process the first entry in the file.  This is handled in 
//  in the main dictionaries by filling the dictionary buffer with '\n'
//  characters before loading short dictionary entries.
//
//
#include "jfc_main.h"
#include "jfc_conf.h"
#include "jfc_dict.h"
#include "jfc_font.h"
#include "jfc_misc.h"

//===================================================================
//
//  Compile time options.
//
#define MAX_KEY_LENGTH  100 // Determines max number of character we allow 
                            //   a search on. If the user search string is 
                            //   too long, we truncate to this level.

//-------------------------------------------------------------------
//
//  Compile time options.
//
#define NAME_MAINDICT       TEXT("edict")       // Name of dictionary file.
#define NAME_USERDICT       TEXT("user.dct")    // Name of user dictionary file.

#define SIZE_LINE       512                     // Offest into the buffer for the search point.
#define SIZE_DICTBUFFER 1024                    // Size of the buffer read from the dicitonary.

//-------------------------------------------------------------------
//
//  static data and definitions.
//

//
//  These variables were moved here from the class because the class item
//  is now allocated on the fly, thus to allow these variables to be 
//  used over more than one instance.  Using this method, the these 
//  files can be loaded the first time the dictionary is used and remain
//  in memory until the system is shutdown.
//
static byte *user_dict    = NULL;   // This is the user dictionary.  This is the 
                                    //   user's own private dictionary.

//-------------------------------------------------------------------
//
//  static routines.
//
static int   dict_comp   (byte *key,byte *ptr,int n);   // Compare a key against a dictionary entry.
static byte *load_dict   (TCHAR *name);                 // Make memory image of a dictionary file.
static int   test_key    (byte *ptr,char *key);         // Test a string against being a dictionary type key.

//
//  Copare a dictionary entry against a key.
//
//      key    -- Key to be compared agianst (ascii should be in lower case,
//                and kana should be converted hiragana!).
//      ptr    -- Pointer to data from the dictionary file.
//      n      -- Length limit.  A value of 0 can be used to unlimit the 
//                comparison (go unitl NULL is reached).
//
//      RETURN -- Zero indicates the key and data match.  A non-zero value
//                indicates they don't match.  Positive indicates key is 
//                later than the data.  Netgitive indicates oposite.
//
static int dict_comp (byte *key,byte *ptr,int n) {
  int  i;
  byte c1,c2;
//  BOOL f1 = false;                              // Flags indicate the first byte of a JIS character code
  BOOL f2 = false;                   

  for (i = 0; ; i++) {
    if (n && (i >= n)) return (0);              // If using limted compare, we can stop when we readch the full length of the string.
    c1 = key[i];
    c2 = ptr[i];
//  if (('A' <= c1) && (c1 <= 'Z')) c1 += 32;   // Key is alraedy in lower case.      
    if (('A' <= c2) && (c2 <= 'Z')) c2 += 32;
//    if (!(c1 & 0x80)) f1 = false;             // Code no longer used because the key is converted to
//      else {                                  //   hiragana in advance!
//        f1 = !f1;
//        if (f1 && (c1 == 0xa5)) c1 = 0xa4;
//    }
    if (!(c2 & 0x80)) f2 = false;
      else {
        f2 = !f2;
        if (f2 && (c2 == 0xa5)) c2 = 0xa4;      // Katakana -> hiragana
      }
    if (c1 != c2) return (c1 - c2);
    if (c1 == 0) return (0);
  }
  return (c1 - c2);
}

//
//  Utility routine to load a dictionary file from disk if the file
//  is available.  The routine does not deal with errors.
//
//      name -- Name of file to load.
//
static byte *load_dict (TCHAR *name) {
  int    i;
  byte  *dict;
  HANDLE handle;
  unsigned long done;

  handle = CreateFile(name,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  if (handle == INVALID_HANDLE_VALUE) return (NULL);
  i = GetFileSize(handle,NULL);
  if (dict = (byte *) calloc(1,i+5)) {
    dict[0] = '\n';                          // Make it look like there is an earlier 
    ReadFile(handle,dict+1,i,&done,NULL);    //   entry (so we can search to beginning)
  }
  CloseHandle (handle);
  return (dict);
}

//
//  This is a utility routien used by the entry type filtering in the 
//  JFC_dict::search_dict() routine.  This routine attempts to see if 
//  the string pointed to matches a articular dictionary type key.  The
//  conditions for such a match are to match the charcters in the key, 
//  and be followed by a ',' or a ')'.
//
//  Note the first character of the input string is skiped, because it
//  is assumed that this is either '(', or ','.
//
//      ptr -- String to be tested.
//      key -- Key to test against (null termianted ascii string).
//
static int test_key (byte *ptr,char *key) {
  int len = strlen(key);
  if (strncmp((char *) ++ptr,key,len)) return (0);
  if ((ptr[len] == ',') || (ptr[len] == ')')) return (len+1);
  return (0);
}

//-------------------------------------------------------------------
//
//  Exported routines.
//

//
//  This is a rotuine called during clean up to deallocate memory helded
//  by the dictionary routines.
//
void free_dictionary () {
  if (user_dict) free (user_dict);          // User dictionary.
  return;
}

//-------------------------------------------------------------------
//
//  begin class JFC_dict
//
//  This class implements the dictionary search.
//

JFC_dict jfc_dict;          // Class instance.

//
//  This routine is called when an entry in the dictionary has been 
//  found.  This rotuine will check the entry against the filters, 
//  and if the entry passess all filters, the entry will be put into
//  the list.
//
//      ptr    -- Pointer to the actual entry location in the dictionary
//                buffer.
//      length -- Length of user's search key.
//
#define NUMBER_DICTNAME (sizeof(names)/sizeof(char [6]))

#define IS_BEGIN(c) ((c == '[') || (c == ' ') || (c == '/') || ISCRLF(c))
#define IS_END(c)   ((c == ']') || (c == ' ') || (c == '/'))

void JFC_dict::check_entry (byte *ptr,int length) {
//
//  Check entry for valid begin/end requirements.
//
  if (!IS_BEGIN(ptr[-1]) || !IS_END(ptr[length])) return;
//
//  Hooray! We have an entry that we actually want to keep.  This means 
//  we have to format the line and generate the output.  Most of the 
//  work in generating the output string is done by the class EUC_buffer.
//  (see top of file for class information).
//                                  
  matches++;                        // Change count and display
  format_line (ptr);                // Output line.
  return;
}

//
//  This routine actually does a search.  The argumetns inciate what 
//  to be searched for.  In a simple search, this will be called once,
//  For an advanced search, this can be called many different times as
//  the search arguments is modified and many different endings are 
//  added.
//
//      search -- The string to be searched for.
//      length -- Length of the search string.
//
//      RETURN -- Indicates the search status:
//
//          DICTSEARCH_MIXED -- Search contains ascii and kana.
//          DICTSEARCH_SHORT -- Too few characters to search for.
//          DICTSEARCH_OKAY  -- Search is OK (this does not mean any
//                              matches were found).
//
int JFC_dict::do_search (KANJI *search,int length) {
  int ascii,count,kanji,kana,i;
  byte key[2*MAX_KEY_LENGTH+4];
//
//  Build the key
//
  ascii = count = kanji = kana = 0;
  for (i = 0; i < length; i++) {                // Build a key.
    if (ISASCII(search[i])) {
       ascii++;
      key[count++] = tolower(search[i]);
    }
    else {
      if (ISKANA(search[i])) kana++; else kanji++;
      key[count++] = HIBYTE(search[i]) | 0x80;
      key[count++] = LOBYTE(search[i]) | 0x80;
    }
  }
  key[count] = 0;
//
//  Check key for validity.
//
  if (ascii && (kana || kanji)) return (DICTSEARCH_MIXED);
  if ((ascii && (ascii < 3)) || (!ascii && !kanji && (kana < 2))) return (DICTSEARCH_SHORT);
//
//  Search user dictionary, and abort point.
//
  if (user_dict) search_memory (key,count,user_dict);
//
//  Next we will setup for the the search loop.  This loop gets each
//  dictionayr entry, checks to see if it still matches the key, then
//  formats the entry for display.
//
//  Search EDICT
//
  if (!matches) search_index (key,count,jfc_config.name(NAME_MAINDICT));
//
//  Done!
//
  return (DICTSEARCH_OKAY);
}

//
//  This routine fromats a dictionary line for display.  This is used
//  by the user dictionary editor and the main dictonary routines.
//
//      data   -- Poiner into the dictionary line (not the beginning 
//                of the line).
//
void JFC_dict::format_line (byte *data) {
  int ch,i,j;
  int first_line = true;
  j = 0;
  for (i = 0; i < SIZE_LINE; i++) {             // Limit string length.
    if (ISCRLF(data[1])) break;                 // End of line so exit.
    else if (*data & 0x80) {                    // Output kanji/kana character.
      ch = *data++;
      output[j++] = (((ch << 8) | *data)) & 0x7f7f;
    }
    else if (first_line) {                      // First line has special characters.
      switch (*data) {
        case '/':                               // First '/' indicates end of first line, just text after here.
             first_line = false;
             break;
        case ']':
        case '[':
        case ' ':
        default:
             output[j++] = (*data);
             break;
      }
    }
    else if (*data == '/') {                    // After first line just output, but change '/' into ', '.
      output[j++] = ',';
      output[j++] = ' ';
    }
    else output[j++] = (*data);
    data++;
  }
  output[j] = 0;
  return;
}

//
//  This routine reads a line from the diction based on an index into 
//  the index file.
//
//      index  -- File handle for index file (should already be open).
//      dict   -- File handle for dictionary file (should already be open).
//      loc    -- Index into dictionary file to get data for.
//      buffer -- Buffer to read line into.  Note that this routine 
//                actually reads data into the middle of the buffer.
//                This is so you can back up in the buffer.  For 
//
//  From search_dict() routine:
//
//  This routine uses buffer and buf, to allow us to backward scan in 
//  the file.  This works as follows:
//
//  When a read from the dictionary takes place, a block of size SIZE_DICTBUFFER
//  bytes is read.  The cener of this block is the locationactually 
//  requested.  This gives us the capability to backup to the beginning 
//  of the dictonary entry.  
//
//  The pointer buf, points into the buffer and points the actual point
//  in the line that was being requested.
//
void JFC_dict::get_line (HANDLE index,HANDLE dict,int loc,byte *buffer) {
  long offset;
  unsigned long done;
  SetFilePointer (index,(loc+1)*sizeof(long),NULL,FILE_BEGIN);
  ReadFile (index,&offset,sizeof(long),&done,NULL);
  if (offset-SIZE_LINE < 0) {                       // Special case for near the beginning of the file.
    memset         (buffer,'\n',SIZE_LINE);         // This is required to prevent search errors.
    SetFilePointer (dict,0,NULL,FILE_BEGIN);
    ReadFile       (dict,buffer+SIZE_LINE-offset,SIZE_DICTBUFFER-SIZE_LINE+offset,&done,NULL);
  }
  else {                                            // General case.
    SetFilePointer (dict,offset-SIZE_LINE,NULL,FILE_BEGIN);
    ReadFile       (dict,buffer,SIZE_DICTBUFFER,&done,NULL);
  }
  return; 
}

//
//  This is the search routine for indexed dictionaries.
//
//      key    -- Key to search for.
//      length -- Length of the key
//      name   -- Name of the dictionary (".jdx" is added to get the name of the index);
//
void JFC_dict::search_index (byte *key,int length,TCHAR *name) {
  HANDLE     dict,index;                // Handles for main dictionary file and for index file.
  long       top,bottom,middle,cut;     // Parameter for binary search.
  int        diff;                      // Difference in key comparisons.
  byte       buffer[SIZE_DICTBUFFER];   // Main buffer for reading in data from the dictionary.
  byte      *buf;                       // Pointer to the location in buffer when the user requested data is.
  TCHAR     *index_name;
//
//  This routine uses buffer and buf, to allow us to backward scan in 
//  the file.  This works as follows:
//
//  When a read from the dictionary takes place, a block of size SIZE_DICTBUFFER
//  bytes is read.  The cener of this block is the locationactually 
//  requested.  This gives us the capability to backup to the beginning 
//  of the dictonary entry.  
//
//  The pointer buf, points into the buffer and points the actual point
//  in the line that was being requested.
//
  buf = buffer+SIZE_LINE-1;
//
//  Open dictionary and index.
//
  index_name = (TCHAR *) buffer;
  lstrcpy (index_name,name);
  lstrcat (index_name,TEXT(".jdx"));
  dict  = CreateFile (      name,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  index = CreateFile (index_name,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  if ((dict == INVALID_HANDLE_VALUE) || (index == INVALID_HANDLE_VALUE)) {
    CloseHandle (dict);
    CloseHandle (index);
    return;
  }
//
//  This section performs a binary search, looking for the beginning
//  of the index region dealing with the string that the user has entered.
//
  top    = 0;
  bottom = (GetFileSize(index,NULL)/sizeof(long))-1;
  while (true) {
    middle=(top+bottom)/2;
    get_line (index,dict,middle,buffer);
    diff = dict_comp (key,buf,0);
    if (!diff) { cut = middle; break; }
    if (top >= bottom-1) { cut = bottom; break; }
    if (diff > 0) top = middle; else bottom = middle;
  }
//
//  The actual loop.
//
  for (;; cut++) {
//
//  Get the next line from the dictionary.  
//
    get_line (index,dict,cut,buffer);
    if (dict_comp(key,buf,length)) break;   // No longer match.  Search is done, let's get out of here.
    check_entry (buf,length);
    if (matches) return;
  }
//
//  Clean up and exit.
//
  CloseHandle (dict);   
  CloseHandle (index);
  return;
}

//
//  This is the main search engine.  This is where you go when you
//  click SEARCH.  This handles getting the parameters from the dialog
//  box.  This routine could have been embedded (and it once was) in 
//  the dialog box handler, however, that is just too hard to deal with,
//  so I sepated it.
//
//  Special macros used to modify the search string then actually do the 
//  search.  These are simply defined to simply the writing of the code.
//  
//      SEARCH_END -- Changes the end character and then does a search.
//      SEARCH_ADD -- Adds a character to the end of the stirng and searches.
//
#define SEARCH_END(x) { ch[0] = x; if (abort = do_search(search,length  )) break; }
#define SEARCH_ADD(x) { ch[1] = x; if (abort = do_search(search,length+1)) break; }

int JFC_dict::search_dict (KANJI *search_ptr,KANJI *result) {
  int    ascii,i,length;
//KANJI *search_ptr;                // Pointer to edit-box version of the search string (NO NOT CHANGE!)
  KANJI  search[MAX_KEY_LENGTH+1];  // Copy of the edit-box search string that can be modified.
//
//  Intitlaize the search parameters.
//
  length = kstrlen(search_ptr);
  if (length > MAX_KEY_LENGTH) length = MAX_KEY_LENGTH;     // Truncate user string if necessary.
  while (length && ISSPACE(search_ptr[length-1])) length--;
//
//  Pre-rpocess key.  This converts the key to hiragana for faster searching,
//  and checks various parameters in the key for later use!
//
  ascii = false;
  for (i = 0; i < length; i++) {
    search[i] = search_ptr[i];
    if      (ISASCII   (search[i])) ascii = true;
    else if (ISKATAKANA(search[i])) search[i] = BASE_HIRAGANA | (search[i] & 0xff);
  }
//
//  Initialize the search system.
//
  matches = 0;
  output  = result;
//
//  Do the search.
//
  do_search (search,length);
  return (matches);
}

//
//  Performs a dictionary search based on a memory dictionary.  The
//  dictionary should be loaded before calling this routine.
//
//      key    -- Key to be searched for.
//      length -- Length of the key.
//      dict   -- Pointer to the dictionary memory image (has a fake
//                '\n' in the first slot).
//
void JFC_dict::search_memory (byte *key,int length,byte *dict) {
  int   i;
  byte *ptr,*p,*p2,buffer[SIZE_BUFFER];

  for (ptr = dict+1; *ptr; ptr++) {
    if (dict_comp(key,ptr,length)) continue;    // No match.
    for (i = 0, p = ptr; (i < SIZE_LINE-9) && !ISCRLF(*ptr); ptr--, i++); 
                                                // Find beginning of entry.
    buffer[0] = '\n';                           // Build duplicate entry.
    p2 = buffer+(p-ptr);                        // Calculate same relative place for end/begin 
    for (i = 1, ptr++; (i < SIZE_BUFFER-2) && !ISCRLF(*ptr); ptr++) buffer[i++] = *ptr;     
                                                // Diplicate entry (use ptr so skip rest of entry)
    buffer[i] = '\n';                           // Terminate enry.
    check_entry (p2,length);                    // Process entry.
    if (matches) return;
  }
  return;
}

//
//  End Class JFC_dict.
//
//-------------------------------------------------------------------








