//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//                                                                   //
//  This module manages the actual card list as a file class object.
//

#ifndef jfc_file_h
#define jfc_file_h

#define IDM_FILE_FILES_BASE     20000   // ID for recent files list.

                                // Card flags
#define CARD_HASKANJI   0x01    // Card has kanji (ie not just kana)
#define CARD_USEKANJI   0x02    // User knows the kanji so make a kanji based card.
#define CARD_KANAONLY   0x04    // Generate a kana only card no matter what.
#define CARD_KANJIONLY  0x08    // Generate a kanji only card no matter what.
#define CARD_ASCIIONLY  0x10    // Generate an ascii only card no matter what.
#define CARD_SUBCARD    0x20    // This is a sub-card of a kanji-card.
#define CARD_KANJI      0x40    // Card is a kanji card (special type)

                                // Duplicate card flags.
#define DUPLICATE_ORIGINAL  1   // Original card that has been pushed to the end
#define DUPLICATE_DUPLICATE 2   // An actual duplicate card

                                // Return codes from CARD_data::parse
#define PARSE_OK        0       // The card data is OK, process it.
#define PARSE_SKIP      1       // Skip this.  Indicates a comment card or an error.
#define PARSE_SUBFILE   2       // Sub-file load.  The file name is encoded as JIS in the ascii field.

//
//  Class used to determine the contents of a card.  This class is only used to parse a 
//  standard card into it's parts.
//
class CARD_data {
public:
  KANJI kana [SIZE_BUFFER+1];           // Kana string for card
  KANJI kanji[SIZE_BUFFER+1];           // Kanji string for card
  KANJI ascii[SIZE_BUFFER+1];           // ASCII (English) meaning.
  KANJI kanji_char[2];                  // Actual kanji character for a kanji card.
  byte  kanji_known;                    // User marked card as knowning the kanji (X)
  byte  kanji_only;                     // Forced kanji based card.
  byte  kana_only;                      // Forced kana based card.
  byte  ascii_only;                     // Forced ascii based card.
  byte  kanji_card;                     // This is a kanji card.
  byte  sub_card;                       // Sub-card of a kanji card.
  int  parse (byte *data,int encoding); // The actual parse call.
  void set   (void);                    // Sets the data for viewing in the main window.
};

//
//  This class manages an actuall card.  Note this could actually generate more than one
//  flash card, because of the possibility of the kana->, kanji->, ascii->
//
class JFC_card {
friend class JFC_file;
friend class SUB_file;
public:
private:  
  class JFC_card *next;         // Linke to next card
  byte           *data;         // Pointer to actual data (unparsed for this card).
  byte            flags;        // Flags for this card.
  byte            stat_kanji;   // Status for kanji cards
  byte            stat_kana;    // Status for kana cards
  byte            stat_ascii;   // Status for ascii cards
  byte            encoding;     // Encoding for this data.
};

//
//  This class is the list of user flash cards.  These are the cards actually shown on the 
//  screen.
//
#define FLASH_ASCII         1       // Show the ascii
#define FLASH_KANA          2       // Show the kana
#define FLASH_KANJI         3       // Show the kanji
#define FLASH_KANJIKANJI    4       // Kanji card show kanji
#define FLASH_KANJIKANA     5       // Kanji card show reaidngs.
#define FLASH_KANJIASCII    6       // Kanji card show meaning
#define FLASH_ERROR        10       // Error card used for generating messages

class JFC_flash {
friend class JFC_file;
public:
private:
  class JFC_flash *next;        // Next card in list.
  class JFC_card  *card;        // Pointer to card data.
  byte duplicate;               // Is this a duplicate card.
  byte type;                    // Type of card to generate.
  byte prob;                    // Probability of bing included (0-100).
};

//
//  This class corresponds to a subfile.  A flash card file always loads at least one
//  subfile, itself.  It may load other files also.
//  
#define ENCODE_EUC  0       // File is encoded as EUC
#define ENCODE_UTF8 1       // FIle is encoded as UTF-8

class SUB_file {
friend class JFC_file;
public:
  SUB_file  (void);
  ~SUB_file (void);
  class SUB_file *next;                     // Pointer to next subfile.
  TCHAR          *filename;                 // Name of the subfile loaded.
  int load  (TCHAR *name,int config);       // Actually load a subfile.
private:
  byte           *buffer;                   // This is a memory image of the file.  The cards will point to sections of this.
  int             encoding;                 // File encoding used.
  long            size;                     // Size of the file image in memory.
  class JFC_card *cards;                    // First card contained in the subfile.
  int             count;                    // Number of cards in the subfile.
  class JFC_card *last;                     // Last card contained in the subifle.
  FILETIME time;
  int    add_card     (byte *data);         // Add a card to the list.  Used when reading cards into the file.
  int    save         (void);               // Save the current file
  TCHAR *stat_name    (TCHAR *buffer);      // Get the name for the statistics file.
  int    save_stats   (void);               // Save the status file.
};

//
//  This class manages a file, including reading the file, and generating a clard list
//  for the user.
//
class JFC_file {
friend class JFC_stat;
friend class SUB_file;
public:
  TCHAR *filename;                          // Name of the current file.
  JFC_file (void);                          // Constructor
  ~JFC_file (void);                         // Destructor.
  void do_menu      (int wParam);           // Do menu items assoicated with the file.
  int  close        (int check);            // Close this file.
  int  read         (TCHAR *name);          // Read in a file and start generating cards.
  void reinit       (int check);            // Called to change options settings.
private:
  int              count;                   // Number of subfiles
  class SUB_file  *files;                   // Pointer to list of subfiles.
  class SUB_file  *last;                    // Pointer to last subfile.
  int              card_count;              // Number of data cards loaded.
  class JFC_card  *cards;                   // List of cards.  This is not the cards shown user, but data cards.
  int              flash_count;             // Number of flash cards.
  class JFC_flash *flash;                   // Actual flash-card list.
  int              right;                   // Number of correct cards.
  int              wrong;                   // Number of wrong cards.
  long             size;                    // Size of file (used for saving stats).
  FILETIME         time;                    // File write time (used for saving stats)

  int    add_card     (byte *data);         // Add a card to the list.  Used when reading cards into the file.
  void   do_stats     (int ok);             // Calcuate statistics for this event
  int    flash_add    (class JFC_card *card,int type);    // Add a flash card to the list beginning.
  void   flash_delete (void);               // Delete topmost flash card.
  int    flash_done   (void);               // Check to see if we are done with the cards.
  void   flash_kanji  (void);               // Display kanji card.
  void   flash_next   (void);               // Put up the next flash card.
  int    read_subfile (TCHAR *name);        // Read a subfile;
  int    read_save    (void);               // Attempt to read a saved file.
  int    save         (void);               // Save the current file

  TCHAR *stat_name    (TCHAR *buffer);      // Get the name for the statistics file.
  TCHAR *save_name    (TCHAR *buffer);      // Get the name for the save file.
};

extern JFC_file *jfc_file;                  // Class instance pointer.

extern void recent_files (TCHAR *name);     // Add name to the recent files list.
extern void window_title (void);            // Set the window title.


#endif jfc_file_h

extern int open_file (TCHAR *name);